
/* Ryan Drake (stiletto@mediaone.net)
 *
 * To build:  gcc -o pngtest pngtest.c libpng.a libz.a -lm
 *
 * Note: You need to build libpng and zlib, which can be found
 *       on the Internet in many places.  If anyone can figure
 *       out how to get this to work with dynamic linking, it
 *       would sure cut down on the executable's size!
 */

png.a ../zlib/libz.a -lm

#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <png.h>

/* EXPORTS */
struct empeg_dev_s {
    int fd;
    caddr_t dmap;
};

struct empeg_bm_s {
    int width;
    int height;
    unsigned char* data;
};

#define PNG_BYTES_TO_CHECK 8
struct empeg_bm_s*
empeg_bmload_png(char* file_name)
{
    png_structp png_ptr;
    png_infop info_ptr;
    png_uint_32 w, h, rowbytes;
    int bit_depth, color_type, row;
    FILE *fp;
    char buf[PNG_BYTES_TO_CHECK];
    png_bytep* row_pointers;

    unsigned char* d;
    unsigned short* sd;
    struct empeg_bm_s* e;
    int i;

    /* Open the prospective PNG file. */
    if ((fp = fopen(file_name, "rb")) == NULL) {
        return 0;
    }

    /* Read in some of the signature bytes */
    if (fread(buf, 1, PNG_BYTES_TO_CHECK, fp) != PNG_BYTES_TO_CHECK) {
        return 0;
    }

    /* Compare the first PNG_BYTES_TO_CHECK bytes of the signature */
    if (png_sig_cmp(buf, (png_size_t)0, PNG_BYTES_TO_CHECK)) {
        return 0;
    }

    /* Create "read struct" */
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (png_ptr == NULL) {
        fclose(fp);
        return 0;
    }

    /* Create "info struct" */
    info_ptr = png_create_info_struct(png_ptr);
    if (info_ptr == NULL) {
        png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
        fclose(fp);
        return 0;
    }

    /* Set error handling */
    if (setjmp(png_ptr->jmpbuf)) {
        png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
        fclose(fp);
        return 0;
    }

    /* Set up the input control */
    png_init_io(png_ptr, fp);

    /* If we have already read some of the signature */
    png_set_sig_bytes(png_ptr, PNG_BYTES_TO_CHECK);

    /* Get info about PNG file */
    png_read_info(png_ptr, info_ptr);
    png_get_IHDR(png_ptr, info_ptr, &w, &h, &bit_depth, &color_type,
         NULL, NULL, NULL);

    /* Expand paletted colors into true RGB triplets */
    if (color_type == PNG_COLOR_TYPE_PALETTE) {
        png_set_expand(png_ptr);
    }

    /* Expand grayscale images to the full 8 bits from 1, 2, or 4 bits/pixel */
    if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8) {
        png_set_expand(png_ptr);
    }

    /* Expand paletted or RGB images with transparency to full alpha channels
     * so the data will be available as RGBA quartets.
     */
    if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS)) {
        png_set_expand(png_ptr);
    }

    /* Allocate the memory to hold the image using the fields of info_ptr. */

    row_pointers = calloc (h, sizeof(*row_pointers));
    if (row_pointers == NULL) {
        png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
        fclose(fp);
        return 0;
    }

    rowbytes = png_get_rowbytes (png_ptr, info_ptr);
    row_pointers[0] = calloc (h, rowbytes);
    if (row_pointers[0] == NULL) {
        free (row_pointers);
        png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
        fclose(fp);
        return 0;
    }

    for (row = 1; row < h; row++) {
        row_pointers[row] = row_pointers[row-1] + rowbytes;
    }

    png_read_image(png_ptr, row_pointers);
    png_read_end(png_ptr, info_ptr);

    /* Set pointer so we can find the data later */
    d = (char*)row_pointers[0];

    /* Clean up */
    free(row_pointers);
    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
    fclose(fp);

    /* Allocate our bitmap structure */
    e = calloc(1, sizeof(*e));
    if(e == NULL) {
        free(d);
        return 0;
    }

    e->width = w;
    e->height = h;

    /* Allocate space for empeg 4bpp bitmap data */
    e->data = malloc (w * h);
    if(e->data == NULL) {
        free(e);
        free(d);
        return 0;
    }

    /* Select format and components */
    switch(color_type) {
        case PNG_COLOR_TYPE_GRAY:

            switch(bit_depth) {
                case 8:
                    memcpy(e->data, d, w * h);
                    break;
                case 16:
                    sd = (unsigned short*)d;
                    i = w * h;
                    while(--i) {
                        e->data[i] = sd[i] >> 8;
                    }
                    break;
                default:
                    /* Unsupported */
                    free(d);
                    return 0;
                    break;
            }
            break;

        case PNG_COLOR_TYPE_GRAY_ALPHA:
        case PNG_COLOR_TYPE_RGB:
        case PNG_COLOR_TYPE_RGB_ALPHA:
        default:
            /* Unsupported */
            free(d);
            return 0;
            break;
    }

    free(d);
    return e;
}

void
empeg_bmfree( struct empeg_bm_s* e )
{
    free(e->data);
    free(e);
}

#define PIXEL(d,x,y) ((d)[ (y) * 64 + ((x) >> 1) ])
#define PIXEL_RIGHT(d,x,y) (PIXEL(d,x,y) & 0xF0)
#define PIXEL_LEFT(d,x,y) (PIXEL(d,x,y) & 0x0F)

static int
empeg_pxread( caddr_t dmap, int x, int y )
{
    if( x & 1) {
        /* right pixel */
        return PIXEL_RIGHT(dmap, x, y) >> 4;
    } else {
        /* left pixel */
        return PIXEL_LEFT(dmap, x, y);
    }
}

/* draw pixel, no alpha, no clamping (v should be >4) */
static void
empeg_pxdraw( caddr_t dmap, int x, int y, int v )
{
    if( x & 1) {
        /* right pixel */
        PIXEL(dmap,x,y) = PIXEL_LEFT(dmap,x,y) | (v<<4);
    } else {
        /* left pixel */
        PIXEL(dmap,x,y) = PIXEL_RIGHT(dmap,x,y) | v;
    }
}

/* draw pixel, blend result = v * alpha + dest * (1-alpha), no clamping */
static void
empeg_pxdraw_a( caddr_t dmap, int x, int y, int v, int a )
{
    int dest, final;

    dest = PIXEL(dmap,x,y);

    if( x & 1) {
        /* right pixel */
        final = ((dest>>4) * (3-a)) + (v * a);
        final /= 3;
        PIXEL(dmap,x,y) = (final<<4) | (dest&0x0F);
    } else {
        /* left pixel */
        final = ((dest&0x0F) * (3-a)) + (v * a);
        final /= 3;
        PIXEL(dmap,x,y) = (final&0x0F) | (dest&0xF0);
    }
}

/* slow full-bitmap draw, arbitrary size, no clipping */
void
empeg_bmdraw( struct empeg_dev_s* empeg, struct empeg_bm_s* bm, int x, int y )
{
    int i, j, k;

    /* Blit the bitmap into the mmapped area */
    k = 0;
    for(j=y; j<bm->height+y; j++) {
        for(i=x; i<bm->width+x; i++) {
            empeg_pxdraw(empeg->dmap, i, j, bm->data[k++]>>6);
        }
    }
}

/* slow full-bitmap draw, arbitrary size, alpha blending, no clipping */
void
empeg_bmdraw_a( struct empeg_dev_s* empeg, struct empeg_bm_s* bm, int x, int y, int a )
{
    int i, j, k;

    /* Blit the bitmap into the mmapped area */
    k = 0;
    for(j=y; j<bm->height+y; j++) {
        for(i=x; i<bm->width+x; i++) {
            empeg_pxdraw_a(empeg->dmap, i, j, bm->data[k++]>>6, a);
        }
    }
}

/* slow full-bitmap draw, arbitrary size, performs clipping */
void
empeg_bmdraw_clip( struct empeg_dev_s* empeg, struct empeg_bm_s* bm, int x, int y )
{
    int i, j, k;

    /* Blit the bitmap into the mmapped area */
    k = 0;
    for(j=y; j<bm->height+y; j++) {
        for(i=x; i<bm->width+x; i++) {
            if((j>=0) && (j<32) && (i>=0) && (i<128)) {
                empeg_pxdraw(empeg->dmap, i, j, bm->data[k]>>6);
            }
            k++;
        }
    }
}

/* slow full-bitmap draw, arbitrary size, performs clipping, alpha blending */
void
empeg_bmdraw_clip_a( struct empeg_dev_s* empeg, struct empeg_bm_s* bm, int x, int y, int a )
{
    int i, j, k;

    /* Blit the bitmap into the mmapped area */
    k = 0;
    for(j=y; j<bm->height+y; j++) {
        for(i=x; i<bm->width+x; i++) {
            if((j>=0) && (j<32) && (i>=0) && (i<128)) {
                empeg_pxdraw_a(empeg->dmap, i, j, bm->data[k]>>6, a);
            }
            k++;
        }
    }
}

/* clear entire empeg display to zeros */
void
empeg_clear( struct empeg_dev_s* empeg )
{
    memset(empeg->dmap, 0, 32 * 128 / 2);
}

/* update empeg screen */
void
empeg_flush( struct empeg_dev_s* empeg )
{
    ioctl(empeg->fd, _IO('d', 0));
}

/* turn empeg display on or off */
void
empeg_display( struct empeg_dev_s* empeg, int on )
{
    ioctl(empeg->fd, _IOW('d', 1, int), on);
}

/* get a fd and mmap handle for controlling the empeg */
struct empeg_dev_s*
empeg_open(void)
{
    int fd;
    caddr_t dmap;
    struct empeg_dev_s* empeg;

    /* Open and mmap empeg display */
    fd = open("/dev/display", O_RDWR);
    if(fd == -1)
        return 0;

    dmap = mmap(0, 2048, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if (dmap == (caddr_t) -1)
        return 0;

    empeg = malloc(sizeof(*empeg));
    if(empeg == NULL) {
        return 0;
    }

    empeg->fd = fd;
    empeg->dmap = dmap;
    return empeg;
}

void
empeg_close(struct empeg_dev_s* empeg)
{
    int fd;
    caddr_t dmap;

    fd = empeg->fd;
    dmap = empeg->dmap;

    free(empeg);
    munmap(dmap, 2048);
    close(fd);
}

main()
{
    struct empeg_dev_s* empeg;
    struct empeg_bm_s* ebm;
    struct empeg_bm_s* pbm;
    int alpha[] = { 0, 1, 2, 3, 3, 2, 1, 0 };
    int i;

    empeg = empeg_open();

    empeg_clear(empeg);
    empeg_flush(empeg);
    empeg_display(empeg, 1);

    ebm = empeg_bmload_png("empeg.png");
    pbm = empeg_bmload_png("pngtest.png");

    empeg_bmdraw(empeg, ebm, 0, 0);
    empeg_flush(empeg);
    sleep(1);

    empeg_bmdraw(empeg, pbm, 0, 0);
    empeg_flush(empeg);
    sleep(1);

    empeg_clear(empeg);
    empeg_flush(empeg);
    sleep(1);

    for(i=0;i<40;i++) {
        empeg_clear(empeg);
        empeg_bmdraw(empeg, ebm, 0, 0);
        empeg_bmdraw_a(empeg, pbm, 0, 0, alpha[i%8]);
        empeg_flush(empeg);
        usleep(125000);
    }

    empeg_bmfree(ebm);
    empeg_bmfree(pbm);

    empeg_display(empeg, 0);

    empeg_close(empeg);
    return 0;
}
